VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "DirectSoundOut"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements IDXCallback

Public Enum DSO_STATUS
    DSO_PLAYING = 0
    DSO_PAUSING
    DSO_STOPPED
End Enum

Private Const BUFFER_MS_MIN     As Long = 500
Private Const BUFFER_MS_MAX     As Long = 8000

Private DXBuffer        As DirectSoundSecondaryBuffer8

Private WithEvents clsTimerEOS  As clsTimer
Attribute clsTimerEOS.VB_VarHelpID = -1

Private udtNotify()     As DSBPOSITIONNOTIFY
Private udtFormat       As DSBUFFERDESC

Private lngSegmentSize  As Long
Private lngBufferSize   As Long
Private lngBufferMS     As Long
Private lngBufferPos    As Long

Private lngCurrentMS    As Long

' Events
Private lngEventSeg1    As Long
Private lngEventSeg2    As Long
Private lngEventMSCnt   As Long

Private clsDataCB       As IDataCallback

Private udeStat         As DSO_STATUS

Private btBuffer()      As Byte

Private blnReady        As Boolean

Private blnEOS          As Boolean
Private lngExpectedEnd  As Long

Private lngLastSeg      As Long

Public Sub DeInitMe()
    If Not lngEventMSCnt = 0 Then
        SetStatus DSO_STOPPED

        clsTimerEOS.Enabled = False

        If lngEventMSCnt > 0 Then _
            DirectX.DestroyEvent lngEventMSCnt
        If lngEventSeg1 > 0 Then _
            DirectX.DestroyEvent lngEventSeg1
        If lngEventSeg2 > 0 Then
            DirectX.DestroyEvent lngEventSeg2
        End If

        lngEventMSCnt = 0
        lngEventSeg1 = 0
        lngEventSeg2 = 0
    End If
End Sub

Public Function InitMe( _
    ByVal samplerate As Long, _
    ByVal Channels As Integer, _
    ByVal BitsPerSample As Integer, _
    datacallback As IDataCallback, _
    Optional ByVal buffer_ms As Long = 2000 _
) As Boolean

    blnReady = False

    ' DeInitMe

    If Not DirectSoundReady Then
        Exit Function
    End If

    If buffer_ms < BUFFER_MS_MIN Then
        Exit Function
    End If

    If buffer_ms > BUFFER_MS_MAX Then
        Exit Function
    End If

    lngBufferMS = buffer_ms

    With udtFormat
        With .fxFormat
            .nSize = LenB(udtFormat.fxFormat)
            .nFormatTag = WAVE_FORMAT_PCM
            .nChannels = Channels
            .lSamplesPerSec = samplerate
            .nBitsPerSample = BitsPerSample
            .nBlockAlign = .nChannels * (.nBitsPerSample / 8)
            .lAvgBytesPerSec = .nBlockAlign * .lSamplesPerSec
        End With

        .lBufferBytes = MSToBytes(lngBufferMS)

        .lFlags = DSBCAPS_CTRLVOLUME Or _
                  DSBCAPS_CTRLPAN Or _
                  DSBCAPS_CTRLPOSITIONNOTIFY Or _
                  DSBCAPS_GETCURRENTPOSITION2 Or _
                  DSBCAPS_GLOBALFOCUS
    End With

    Set DXBuffer = DirectSound.CreateSoundBuffer(udtFormat)

    lngBufferSize = MSToBytes(lngBufferMS)
    lngSegmentSize = lngBufferSize / 2

    If lngEventMSCnt = 0 Then _
        lngEventMSCnt = frmDXCallback.CreateEvent()
    If lngEventSeg1 = 0 Then _
        lngEventSeg1 = frmDXCallback.CreateEvent()
    If lngEventSeg2 = 0 Then
        lngEventSeg2 = frmDXCallback.CreateEvent()
    End If

    ReDim udtNotify(2) As DSBPOSITIONNOTIFY

    With udtNotify(0)
        .hEventNotify = lngEventSeg1
        .lOffset = (lngBufferSize / 2) - (lngSegmentSize / 2)
    End With

    With udtNotify(1)
        .hEventNotify = lngEventSeg2
        .lOffset = lngBufferSize - (lngSegmentSize / 2)
    End With

    With udtNotify(2)
        .hEventNotify = lngEventMSCnt
        .lOffset = lngBufferSize - 5
    End With

    DXBuffer.SetNotificationPositions UBound(udtNotify) + 1, _
                                      udtNotify

    Set frmDXCallback.clsCallback = Me

    lngCurrentMS = 0
    lngExpectedEnd = -1
    lngBufferPos = 0
    lngLastSeg = 2

    ReDim btBuffer(lngBufferSize - 1) As Byte

    Set clsDataCB = datacallback

    blnEOS = False
    blnReady = True
    InitMe = True
End Function

Public Property Get position( _
) As Long

    Dim udtPos  As DSCURSORS

    If Not blnReady Then Exit Property

    DXBuffer.GetCurrentPosition udtPos

    position = lngCurrentMS + BytesToMS(udtPos.lPlay)
End Property

Public Property Let position( _
    ByVal lngPos As Long _
)

    lngCurrentMS = lngPos - (lngPos Mod lngBufferMS)
End Property

Public Property Get OutSamplerate( _
) As Long

    OutSamplerate = udtFormat.fxFormat.lSamplesPerSec
End Property

Public Property Get OutChannels( _
) As Integer

    OutChannels = udtFormat.fxFormat.nChannels
End Property

Public Property Get OutBitsPerSample( _
) As Integer

    OutBitsPerSample = udtFormat.fxFormat.nBitsPerSample
End Property

Public Function CaptureSamples( _
    ByVal buf_ptr As Long, _
    ByVal buf_len As Long, _
    ByRef buf_written As Long _
) As Boolean

    If Not blnReady Then Exit Function

    Dim udtPos      As DSCURSORS
    Dim lngPNext    As Long
    Dim lngCapStart As Long

    If Status <> DSO_PLAYING Then
        Exit Function
    End If

    If buf_len > lngBufferSize Then
        Exit Function
    End If

    DXBuffer.GetCurrentPosition udtPos

    lngCapStart = udtPos.lPlay

    If lngCapStart + buf_len > lngBufferSize Then
        DXCopyMemory ByVal buf_ptr, _
                     btBuffer(lngCapStart), _
                     lngBufferSize - lngCapStart

        buf_written = buf_written + (lngBufferSize - lngCapStart)
        lngPNext = lngBufferSize - lngCapStart
        buf_len = buf_len - (lngBufferSize - lngCapStart)

        DXCopyMemory ByVal buf_ptr + lngPNext, _
                     btBuffer(0), _
                     buf_len

        buf_written = buf_written + buf_len
    Else
        DXCopyMemory ByVal buf_ptr, _
                     btBuffer(lngCapStart), _
                     buf_len

        buf_written = buf_written + buf_len
    End If

    CaptureSamples = True
End Function

Public Sub ResetBuffer()
    DXBuffer.Stop
    udeStat = DSO_STOPPED
    SetStatus DSO_PLAYING
End Sub

Public Function SetStatus( _
    ByVal stat As DSO_STATUS _
) As Boolean

    If Not blnReady Then Exit Function

    Select Case stat
        Case DSO_PAUSING
            If Status = DSO_PAUSING Then
                DXBuffer.Play DSBPLAY_LOOPING
                Status = DSO_PLAYING
                SetStatus = True
                Exit Function
            End If

            DXBuffer.Stop
            Status = DSO_PAUSING
            SetStatus = True

        Case DSO_STOPPED
            DXBuffer.Stop
            DXBuffer.SetCurrentPosition 0

            lngBufferPos = 0
            lngCurrentMS = 0

            Status = DSO_STOPPED
            SetStatus = True

        Case DSO_PLAYING
            If Status = DSO_PLAYING Then
                SetStatus = True
                Exit Function
            ElseIf Status = DSO_PAUSING Then
                DXBuffer.Play DSBPLAY_LOOPING
                Status = DSO_PLAYING
                SetStatus = True
                Exit Function
            End If

            GetAndWriteSegment 1
            lngLastSeg = 1

            DXBuffer.Play DSBPLAY_LOOPING
            Status = DSO_PLAYING
            SetStatus = True

    End Select
End Function

Private Sub GetAndWriteSegment( _
    Optional segment As Long = 0 _
)

    Dim lngWritten  As Long

    If blnEOS Then
        DXBuffer.Stop
        DXBuffer.SetCurrentPosition 0
        Status = DSO_STOPPED
        lngCurrentMS = 0
        lngLastSeg = 0
        Exit Sub
    End If

    If segment > 0 Then
        If segment = 1 Then
            lngBufferPos = 0
        Else
            lngBufferPos = lngBufferSize / 2
        End If
    End If

    clsDataCB.NeedData VarPtr(btBuffer(lngBufferPos)), _
                       lngSegmentSize, _
                       lngWritten

    If lngWritten < lngSegmentSize Then
        blnEOS = True
        lngExpectedEnd = lngCurrentMS + MSToBytes(lngBufferPos + lngWritten)

        If udtFormat.fxFormat.nBitsPerSample = 16 Then
            FillMem btBuffer(lngBufferPos + lngWritten), _
                    lngSegmentSize - lngWritten, _
                    0
        Else
            FillMem btBuffer(lngBufferPos + lngWritten), _
                    lngSegmentSize - lngWritten, _
                    128
        End If
    End If

    WriteToBuffer VarPtr(btBuffer(lngBufferPos)), _
                  lngSegmentSize
End Sub

Private Sub WriteToBuffer( _
    ByVal p As Long, _
    ByVal Length As Long _
)

    DXBuffer.WriteBuffer lngBufferPos, _
                         Length, _
                         ByVal p, _
                         DSBLOCK_DEFAULT

    lngBufferPos = lngBufferPos + Length
    If lngBufferPos >= lngBufferSize Then
        lngBufferPos = 0
    End If
End Sub

Public Property Get Status( _
) As DSO_STATUS

    Status = udeStat
End Property

Private Property Let Status( _
    ByVal ude As DSO_STATUS _
)

    udeStat = ude

    If ObjPtr(clsDataCB) > 0 Then
        clsDataCB.StatusChanged udeStat
    End If
End Property

Public Property Get BufferSizeMS( _
) As Long

    BufferSizeMS = lngBufferMS
End Property

Public Property Get BufferSize( _
) As Long

    BufferSize = lngBufferSize
End Property

Public Property Get Volume( _
) As Long

    If Not blnReady Then Exit Property

    Volume = DXBuffer.GetVolume
End Property

Public Property Let Volume( _
    ByVal lngVal As Long _
)

    If Not blnReady Then Exit Property

    DXBuffer.SetVolume lngVal
End Property

Public Property Get Panning( _
) As Long

    If Not blnReady Then Exit Property

    Panning = DXBuffer.GetPan
End Property

Public Property Let Panning( _
    ByVal lngVal As Long _
)

    If Not blnReady Then Exit Property

    DXBuffer.SetPan lngVal
End Property

Private Sub Class_Initialize()
    lngBufferMS = 2000

    Status = DSO_STOPPED

    Set clsTimerEOS = New clsTimer
    clsTimerEOS.Interval = 50
    clsTimerEOS.Enabled = True

    lngExpectedEnd = -1

    If Not DirectSoundReady() Then
        InitDirectSound 44100, 2, 16
    End If
End Sub

Private Sub Class_Terminate()
    Set DXBuffer = Nothing

    clsTimerEOS.Enabled = False

    If lngEventMSCnt > 0 Then _
        DirectX.DestroyEvent lngEventMSCnt
    If lngEventSeg1 > 0 Then _
        DirectX.DestroyEvent lngEventSeg1
    If lngEventSeg2 > 0 Then
        DirectX.DestroyEvent lngEventSeg2
    End If
End Sub

Private Sub clsTimerEOS_Timer()
    If blnEOS Then
        If Status = DSO_STOPPED Then

            clsDataCB.EndOfStream
            blnEOS = False

        ElseIf Status = DSO_PLAYING Then

            If position >= lngExpectedEnd Then
                DXBuffer.Stop
                DXBuffer.SetCurrentPosition 0

                Status = DSO_STOPPED

                lngCurrentMS = 0
                lngLastSeg = 0

                lngExpectedEnd = -1

                blnEOS = False
            End If

        End If
    End If
End Sub

Private Property Get IDXCallback_MyEvents( _
) As Long()

    Dim lngEvents(2)    As Long

    lngEvents(0) = lngEventMSCnt
    lngEvents(1) = lngEventSeg1
    lngEvents(2) = lngEventSeg2

    IDXCallback_MyEvents = lngEvents
End Property

Private Sub IDXCallback_OnEvent( _
    ByVal eventid As Long _
)

    Select Case eventid
        Case lngEventMSCnt
            lngCurrentMS = lngCurrentMS + lngBufferMS
        Case lngEventSeg1
            If lngLastSeg = 1 Then
                GetAndWriteSegment 2
                lngLastSeg = 2
            End If
        Case lngEventSeg2
            If lngLastSeg = 2 Then
                GetAndWriteSegment 1
                lngLastSeg = 1
            End If
    End Select
End Sub

Private Function BytesToMS( _
    ByVal bytes As Long _
) As Long

    BytesToMS = bytes / udtFormat.fxFormat.lAvgBytesPerSec * 1000
End Function

Private Function MSToBytes( _
    ByVal ms As Long _
) As Long

    MSToBytes = ms / 1000 * udtFormat.fxFormat.lAvgBytesPerSec
End Function
