VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "EncoderAPE"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements IEncoder


Private Const Description   As String = "Monkey's Audio"
Private Const Extension     As String = "ape"


Private Declare Function APECompress_Create Lib "MACDll.dll" _
Alias "c_APECompress_Create" ( _
    pErrorCode As Long _
) As Long

Private Declare Sub APECompress_Destroy Lib "MACDll.dll" _
Alias "c_APECompress_Destroy" ( _
    ByVal handle As APE_HANDLE _
)

Private Declare Function APECompress_Start Lib "MACDll.dll" _
Alias "c_APECompress_Start" ( _
    ByVal handle As APE_HANDLE, _
    ByVal pOutputFilename As String, _
    pwfeInput As Any, _
    ByVal nMaxAudioBytes As Long, _
    ByVal nCompressionLevel As Long, _
    ByVal pHeaderData As Long, _
    ByVal nHeaderBytes As Long _
) As Long

Private Declare Function APECompress_AddData Lib "MACDll.dll" _
Alias "c_APECompress_AddData" ( _
    ByVal handle As APE_HANDLE, _
    pData As Any, _
    ByVal nBytes As Long _
) As Long

Private Declare Function APECompress_GetBufferBytesAvailable Lib "MACDll.dll" _
Alias "c_APECompress_GetBufferBytesAvailable" ( _
    ByVal handle As APE_HANDLE _
) As Long

Private Declare Function APECompress_LockBuffer Lib "MACDll.dll" _
Alias "c_APECompress_LockBuffer" ( _
    ByVal handle As APE_HANDLE, _
    pBytesAvailable As Long _
) As Long

Private Declare Function APECompress_UnlockBuffer Lib "MACDll.dll" _
Alias "c_APECompress_UnlockBuffer" ( _
    ByVal handle As APE_HANDLE, _
    ByVal nBytesAdded As Long, _
    ByVal bProcess As Long _
) As Long

Private Declare Function APECompress_Finish Lib "MACDll.dll" _
Alias "c_APECompress_Finish" ( _
    ByVal handle As APE_HANDLE, _
    ByVal pTerminatingData As Long, _
    ByVal nTerminatingBytes As Long, _
    ByVal nWAVTerminatingBytes As Long _
) As Long

Private Declare Function APECompress_Kill Lib "MACDll.dll" _
Alias "c_APECompress_Kill" ( _
    ByVal handle As APE_HANDLE _
) As Long


Private Type WAVEFORMATEX
    wFormatTag      As Integer
    nChannels       As Integer
    nSamplesPerSec  As Long
    nAvgBytesPerSec As Long
    nBlockAlign     As Integer
    wBitsPerSample  As Integer
    cbSize          As Integer
End Type

Private Type APE_TAG_FOOTER
    preample        As String * 8
    tagver          As Long
    tagsize         As Long
    cTags           As Long
    flags           As Long
    zero1           As Long
    zero2           As Long
End Type

Private Type APE_TAG_ITEM
    size            As Long
    flags           As Long
End Type

Public Enum APE_COMPRESSION_LEVELS
    COMPRESSION_LEVEL_FAST = 1000
    COMPRESSION_LEVEL_NORMAL = 2000
    COMPRESSION_LEVEL_HIGH = 3000
    COMPRESSION_LEVEL_EXTRA_HIGH = 4000
    COMPRESSION_LEVEL_INSANE = 5000
End Enum

Private Enum APE_HANDLE
    INVALID_HANDLE = 0
End Enum

Private Const WAVE_FORMAT_PCM                       As Long = 1&
Private Const MAX_AUDIO_BYTES_UNKNOWN               As Long = -1&
Private Const CREATE_WAV_HEADER_ON_DECOMPRESSION    As Long = -1&


Private hAPE            As APE_HANDLE
Private udeCompLevel    As APE_COMPRESSION_LEVELS

Private wfx_out         As WAVEFORMATEX

Private strOutput       As String

Private clsTags         As StreamTags

Private Sub Class_Initialize()
    hAPE = INVALID_HANDLE
    udeCompLevel = COMPRESSION_LEVEL_NORMAL
End Sub

Public Property Get Quality( _
) As APE_COMPRESSION_LEVELS

    Quality = udeCompLevel
End Property

Public Property Let Quality( _
    ByVal udeQ As APE_COMPRESSION_LEVELS _
)

    udeCompLevel = udeQ
End Property

Private Property Let IEncoder_ConfigString( _
    ByVal RHS As String _
)

    Dim strParts()  As String

    If InStr(RHS, " ") < 1 Then Exit Property

    strParts = Split(RHS, " ")
    If strParts(0) <> "APECFG" Then Exit Property

    udeCompLevel = Val(strParts(1))
End Property

Private Property Get IEncoder_ConfigString( _
) As String

    IEncoder_ConfigString = "APECFG " & udeCompLevel
End Property

Private Property Get IEncoder_Description( _
) As String

    IEncoder_Description = Description
End Property

Private Property Get IEncoder_Extension( _
) As String

    IEncoder_Extension = Extension
End Property

Private Function IEncoder_Encode( _
    ByVal buffer_ptr As Long, _
    ByVal buffer_len As Long, _
    buffer_used As Long _
) As SND_RESULT

    Dim lngRet  As Long

    If buffer_len > IEncoder_EncoderMetrics(ENC_M_PCM_PER_DELIVERY_MAX) Then
        IEncoder_Encode = SND_ERR_INVALID_BUFFER
        Exit Function
    ElseIf buffer_len < IEncoder_EncoderMetrics(ENC_M_PCM_PER_DELIVERY_MIN) Then
        IEncoder_Encode = SND_ERR_INVALID_BUFFER
        Exit Function
    End If

    lngRet = APECompress_AddData(hAPE, _
                                 ByVal buffer_ptr, _
                                 buffer_len)

    If lngRet = 0 Then
        buffer_used = buffer_len
        IEncoder_Encode = SND_ERR_SUCCESS
    Else
        buffer_used = 0
        IEncoder_Encode = SND_ERR_WRITE_ERROR
    End If
End Function

Private Function IEncoder_EncoderClose( _
) As SND_RESULT

    Dim lngRet  As Long

    lngRet = APECompress_Finish(hAPE, 0, 0, 0)
    APECompress_Destroy hAPE
    hAPE = INVALID_HANDLE

    WriteAPETags strOutput, clsTags
    Set clsTags = Nothing

    IEncoder_EncoderClose = SND_ERR_SUCCESS
End Function

Private Function IEncoder_EncoderInit( _
    ByVal samplerate As Long, _
    ByVal Channels As Integer, _
    ByVal output As String, _
    tags As StreamTags _
) As SND_RESULT

    Dim lngRet  As Long

    IEncoder_EncoderInit = SND_ERR_INVALID_PARAM

    If samplerate < 2 Then Exit Function
    If Channels < 1 Or Channels > 2 Then Exit Function
    If output = "" Then Exit Function

    IEncoder_EncoderInit = SND_ERR_INTERNAL

    ' check if monkey's audio codec is available
    If Not IsValidFile("MACDll.dll") Then
        If Not IsValidFile(GetSystemDirectory & "MACDll.dll") Then
            Exit Function
        End If
    End If

    With wfx_out
        .wFormatTag = WAVE_FORMAT_PCM
        .nSamplesPerSec = samplerate
        .nChannels = Channels
        .wBitsPerSample = 16
        .nBlockAlign = .nChannels * (.wBitsPerSample / 8)
        .nAvgBytesPerSec = .nSamplesPerSec * .nBlockAlign
    End With

    hAPE = APECompress_Create(lngRet)
    If lngRet <> 0 Then
        Exit Function
    End If

    lngRet = APECompress_Start(hAPE, _
                               output, _
                               wfx_out, _
                               MAX_AUDIO_BYTES_UNKNOWN, _
                               udeCompLevel, _
                               0, _
                               CREATE_WAV_HEADER_ON_DECOMPRESSION)

    If lngRet = 0 Then
        IEncoder_EncoderInit = SND_ERR_SUCCESS

        ' APE Tags get written after encoding,
        ' so save a reference
        Set clsTags = tags

        strOutput = output
    Else
        APECompress_Destroy hAPE
        hAPE = INVALID_HANDLE
    End If
End Function

Private Function IEncoder_EncoderMetrics( _
    ByVal info As ENCODER_METRICS _
) As Long

    Select Case info
        Case ENC_M_PCM_PER_DELIVERY_MAX
            IEncoder_EncoderMetrics = wfx_out.nAvgBytesPerSec
        Case ENC_M_PCM_PER_DELIVERY_MIN
            IEncoder_EncoderMetrics = wfx_out.nBlockAlign
        Case ENC_M_SUPPORT_TAGS
            IEncoder_EncoderMetrics = ENC_TAG_ALBUM Or _
                                      ENC_TAG_ARTIST Or _
                                      ENC_TAG_COMMENT Or _
                                      ENC_TAG_GENRE Or _
                                      ENC_TAG_TITLE Or _
                                      ENC_TAG_TRACK_NUM Or _
                                      ENC_TAG_YEAR
    End Select
End Function

' APE Tags sind UTF-8!
Private Sub WriteAPETags( _
    ByVal FileName As String, _
    tags As StreamTags _
)

    Dim lngTagsSize As Long
    Dim lngTagStart As Long
    Dim cTags       As Long
    Dim clsTag      As StreamTag
    Dim strTagKey   As String
    Dim i           As Long
    Dim udtFooter   As APE_TAG_FOOTER
    Dim udtItem     As APE_TAG_ITEM

    Dim fh          As Integer

    fh = FreeFile
    Open FileName For Binary As #fh
        Seek #fh, LOF(fh)
        lngTagStart = Seek(fh)

        For i = 1 To clsTags.TagCount
            Set clsTag = clsTags.TagItem(i)
            If Not clsTag.TagName = "" Or clsTag.TagValue = "" Then

                If clsTag.TagName = "tracknumber" Then
                    strTagKey = "track"
                Else
                    strTagKey = clsTag.TagName
                End If

                udtItem.size = Len(clsTag.TagValue) + 1
                Put #fh, , udtItem
                Put #fh, , CStr(strTagKey & Chr$(0))
                Put #fh, , CStr(clsTag.TagValue & Chr$(0))
                cTags = cTags + 1

            End If
        Next

        lngTagsSize = Seek(fh) - lngTagStart + Len(udtFooter)

        With udtFooter
            .preample = "APETAGEX"
            .cTags = cTags
            .tagsize = lngTagsSize
            .tagver = 1000
        End With

        Put #fh, , udtFooter
    Close #fh
End Sub
