VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "EncoderOGG"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements IEncoder


Private Const Extension     As String = "ogg"
Private Const Description   As String = "Ogg Vorbis Encoder"

Public Enum ENC_OV_MODE
    OV_ENC_ABR
    OV_ENC_QUALITY
End Enum

Private Type ogg_page
    pHeader                 As Long
    header_len              As Long
    pBody                   As Long
    body_len                As Long
End Type

Private sngQuality          As Single

Private lngBitrateNom       As Long
Private lngBitrateMin       As Long
Private lngBitrateMax       As Long

Private udeMode             As ENC_OV_MODE

Private os                  As memory   ' ogg stream state
Private og                  As memory   ' ogg page
Private op                  As memory   ' ogg packet
Private vc                  As memory   ' vorbis comment
Private vi                  As memory   ' vorbis info
Private vd                  As memory   ' vorbis dsp state
Private vb                  As memory   ' vorbis block

Private clsOgg              As clsDLL
Private clsVorbis           As clsDLL

Private blnReady            As Boolean

Private hFOut               As hFile

Private intEncChannels      As Long
Private lngEncSamplerate    As Long

Private Const STRUCT_SIZE   As Long = 64& * 1024&

'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''

Private Sub Class_Initialize()
    Set clsOgg = New clsDLL
    Set clsVorbis = New clsDLL

    lngBitrateMax = -1
    lngBitrateMin = -1
    lngBitrateNom = 128000

    sngQuality = 0.4

    udeMode = OV_ENC_QUALITY

    blnReady = True

    blnReady = blnReady And clsOgg.LoadDLL("ogg.dll")
    blnReady = blnReady And clsVorbis.LoadDLL("vorbis.dll")

    If Not blnReady Then Exit Sub

    clsOgg.IsCDECL = True
    clsVorbis.IsCDECL = True

    hFOut.handle = INVALID_HANDLE
End Sub

Private Sub Class_Terminate()
    IEncoder_EncoderClose
End Sub

'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''

Public Property Get Mode( _
) As ENC_OV_MODE

    Mode = udeMode
End Property

Public Property Let Mode( _
    ByVal udeVal As ENC_OV_MODE _
)

    udeMode = udeVal
End Property

Public Property Get Quality( _
) As Single

    Quality = sngQuality
End Property

Public Property Let Quality( _
    ByVal sngVal As Single _
)

    sngQuality = sngVal
End Property

Public Property Get BitrateMin( _
) As Long

    BitrateMin = lngBitrateMin
End Property

Public Property Let BitrateMin( _
    ByVal lngVal As Long _
)

    lngBitrateMin = lngVal
End Property

Public Property Get BitrateMax( _
) As Long

    BitrateMax = lngBitrateMax
End Property

Public Property Let BitrateMax( _
    ByVal lngVal As Long _
)

    lngBitrateMax = lngVal
End Property

Public Property Get BitrateNominal( _
) As Long

    BitrateNominal = lngBitrateNom
End Property

Public Property Let BitrateNominal( _
    ByVal lngVal As Long _
)

    lngBitrateNom = lngVal
End Property

'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''

Private Property Let IEncoder_ConfigString( _
    ByVal RHS As String _
)
    
    Dim strParts()  As String

    If InStr(RHS, " ") < 1 Then Exit Property

    strParts = Split(RHS, " ")
    If strParts(0) <> "OGGCFG" Then Exit Property

    lngBitrateMax = CLng(strParts(1))
    lngBitrateMin = CLng(strParts(2))
    lngBitrateNom = CLng(strParts(3))
       sngQuality = CSng(strParts(4))
          udeMode = CLng(strParts(5))
End Property

Private Property Get IEncoder_ConfigString( _
) As String

    IEncoder_ConfigString = "OGGCFG " & lngBitrateMax & " " & _
                                        lngBitrateMin & " " & _
                                        lngBitrateNom & " " & _
                                        sngQuality & " " & _
                                        udeMode
End Property

Private Property Get IEncoder_Description( _
) As String

    IEncoder_Description = Description
End Property

Private Property Get IEncoder_Extension( _
) As String

    IEncoder_Extension = Extension
End Property

Private Function IEncoder_EncoderInit( _
    ByVal samplerate As Long, _
    ByVal Channels As Integer, _
    ByVal output As String, _
    tags As StreamTags _
) As SND_RESULT

    Dim clsTag                          As StreamTag
    Dim lngSerial                       As Long
    Dim lngRet                          As Long
    Dim i                               As Long
    Dim btHeader(STRUCT_SIZE - 1)       As Byte
    Dim btHeaderComm(STRUCT_SIZE - 1)   As Byte
    Dim btHeaderCode(STRUCT_SIZE - 1)   As Byte

    If Not blnReady Then
        IEncoder_EncoderInit = SND_ERR_INTERNAL
        Exit Function
    End If

    IEncoder_EncoderClose

    If Channels > 2 Or Channels < 1 Then Exit Function
    If samplerate < 1 Then Exit Function
    If output = "" Then Exit Function

    hFOut = FileOpen(output, GENERIC_WRITE, , CREATE_ALWAYS)
    If hFOut.handle = INVALID_HANDLE Then
        IEncoder_EncoderInit = SND_ERR_INVALID_OUTPUT
        Exit Function
    End If

    AllocStructs

    ' Vorbis Codec
    vorbis_info_init vi.pointer

    ' Vorbis Bitrate Management
    Select Case True

        Case udeMode = OV_ENC_QUALITY
            lngRet = vorbis_encode_init_vbr(vi.pointer, _
                                            Channels, _
                                            samplerate, _
                                            sngQuality)

        Case udeMode = OV_ENC_ABR
            lngRet = vorbis_encode_init(vi.pointer, _
                                        Channels, _
                                        samplerate, _
                                        lngBitrateMax, _
                                        lngBitrateNom, _
                                        lngBitrateMin)

        Case Else
            lngRet = -1

    End Select

    If lngRet < 0 Then
        IEncoder_EncoderInit = SND_ERR_INVALID_PARAM
        FileClose hFOut
        FreeStructs
        Exit Function
    End If

    vorbis_comment_init vc.pointer

    ' write tags in comment header
    If Not tags Is Nothing Then
        For i = 1 To tags.TagCount

            Set clsTag = tags.TagItem(i)

            Select Case LCase$(clsTag.TagName)

                Case "artist", "album", _
                     "title", "genre", _
                     "year", "comment", _
                     "tracknumber":

                    vorbis_comment_add_tag vc.pointer, _
                                           clsTag.TagName, _
                                           clsTag.TagValue

            End Select

        Next
    End If

    ' Vorbis DSP
    lngRet = vorbis_analysis_init(vd.pointer, vi.pointer)
    If lngRet <> 0 Then
        IEncoder_EncoderInit = SND_ERR_UNKNOWN
        FileClose hFOut
        FreeStructs
        Exit Function
    End If

    lngRet = vorbis_block_init(vd.pointer, vb.pointer)
    If lngRet <> 0 Then
        IEncoder_EncoderInit = SND_ERR_UNKNOWN
        FileClose hFOut
        FreeStructs
        Exit Function
    End If

    ' Ogg Bitstream
    lngSerial = Rnd() * 100000
    '
    lngRet = ogg_stream_init(os.pointer, lngSerial)
    If lngRet <> 0 Then
        IEncoder_EncoderInit = SND_ERR_UNKNOWN
        FileClose hFOut
        FreeStructs
        Exit Function
    End If

    ' Ogg/Vorbis Headers
    vorbis_analysis_headerout vd.pointer, _
                              vc.pointer, _
                              VarPtr(btHeader(0)), _
                              VarPtr(btHeaderComm(0)), _
                              VarPtr(btHeaderCode(0))

    ogg_stream_packetin os.pointer, VarPtr(btHeader(0))
    ogg_stream_packetin os.pointer, VarPtr(btHeaderComm(0))
    ogg_stream_packetin os.pointer, VarPtr(btHeaderCode(0))

    ' flush the 3 headers, so actual data starts
    ' on a new page
    If Not WriteStream(True) Then
        IEncoder_EncoderInit = SND_ERR_WRITE_ERROR
        FileClose hFOut
        FreeStructs
        Exit Function
    End If

    intEncChannels = Channels
    lngEncSamplerate = samplerate

    IEncoder_EncoderInit = SND_ERR_SUCCESS
End Function

Private Function IEncoder_Encode( _
    ByVal buffer_ptr As Long, _
    ByVal buffer_len As Long, _
    buffer_used As Long _
) As SND_RESULT

    Dim intSamples()    As Integer
    Dim sngSamples()    As Single
    Dim cSamples        As Integer
    '
    Dim ppBuffers       As Long
    Dim pBuffers()      As Long
    '
    Dim j               As Long
    Dim i               As Long

    If Not blnReady Then
        IEncoder_Encode = SND_ERR_INTERNAL
        Exit Function
    End If

    If buffer_len > IEncoder_EncoderMetrics(ENC_M_PCM_PER_DELIVERY_MAX) Then
        buffer_len = IEncoder_EncoderMetrics(ENC_M_PCM_PER_DELIVERY_MAX)
    ElseIf buffer_len < IEncoder_EncoderMetrics(ENC_M_PCM_PER_DELIVERY_MIN) Then
        IEncoder_Encode = SND_ERR_INVALID_BUFFER
        Exit Function
    End If

    cSamples = buffer_len / 2

    ReDim intSamples(cSamples - 1) As Integer
    ReDim sngSamples(cSamples - 1) As Single

    ' get a buffer from vorbis encoder for new samples
    ppBuffers = vorbis_analysis_buffer(vd.pointer, _
                                       cSamples / intEncChannels)

    If ppBuffers = 0 Then
        IEncoder_Encode = SND_ERR_INVALID_BUFFER
        Exit Function
    End If

    ' channel buffer array
    ' 1 buffer for each channel
    ReDim pBuffers(intEncChannels - 1) As Long

    CpyMem pBuffers(0), _
           ByVal ppBuffers, _
           intEncChannels * 4

    CpyMem intSamples(0), _
           ByVal buffer_ptr, _
           buffer_len

    ' Deinterleaving
    If intEncChannels = 2 Then
        ' Channel L
        j = 0
        For i = 0 To cSamples / 2 - 1
            sngSamples(i) = intSamples(j) / 32768#
            j = j + 2
        Next

        ' Channel R
        j = 1
        For i = cSamples / 2 To cSamples - 1
            sngSamples(i) = intSamples(j) / 32768#
            j = j + 2
        Next
    Else
        ' Mono
        For i = 0 To cSamples - 1
            sngSamples(i) = intSamples(i) / 32768#
        Next
    End If

    If intEncChannels = 2 Then

        CpyMem ByVal pBuffers(0), _
               sngSamples(0), _
               cSamples / 2 * 4

        CpyMem ByVal pBuffers(1), _
               sngSamples(cSamples / 2), _
               cSamples / 2 * 4

    Else

        CpyMem ByVal pBuffers(0), _
               sngSamples(0), _
               cSamples * 4

    End If

    ' tell vorbis how many samples we wrote
    vorbis_analysis_wrote vd.pointer, _
                          cSamples / intEncChannels

    ' encode written samples and write pages
    ' to the file
    ProcessAudio

    buffer_used = buffer_len

    IEncoder_Encode = SND_ERR_SUCCESS
End Function

Private Function IEncoder_EncoderClose( _
) As SND_RESULT

    If Not blnReady Then
        IEncoder_EncoderClose = SND_ERR_INTERNAL
        Exit Function
    End If

    If hFOut.handle = INVALID_HANDLE Then
        IEncoder_EncoderClose = SND_ERR_INVALID_OUTPUT
        Exit Function
    End If

    ' tell vorbis we're at the end of the stream
    vorbis_analysis_wrote vd.pointer, 0
    ProcessAudio

    ogg_stream_clear os.pointer
    vorbis_block_clear vb.pointer
    vorbis_dsp_clear vd.pointer
    vorbis_comment_clear vc.pointer
    vorbis_info_clear vi.pointer

    FileClose hFOut
    FreeStructs

    lngEncSamplerate = 0
    intEncChannels = 0

    IEncoder_EncoderClose = SND_ERR_SUCCESS
End Function

Private Function IEncoder_EncoderMetrics( _
    ByVal info As ENCODER_METRICS _
) As Long

    Select Case True
        Case info = ENC_M_SUPPORT_TAGS
            IEncoder_EncoderMetrics = ENC_TAG_ALBUM Or _
                                      ENC_TAG_ARTIST Or _
                                      ENC_TAG_GENRE Or _
                                      ENC_TAG_TITLE Or _
                                      ENC_TAG_YEAR Or _
                                      ENC_TAG_COMMENT

        Case info = ENC_M_PCM_PER_DELIVERY_MAX
            ' 1/2 second of data (not tested)
            IEncoder_EncoderMetrics = lngEncSamplerate

        Case info = ENC_M_PCM_PER_DELIVERY_MIN
            ' not tested either
            IEncoder_EncoderMetrics = 4

    End Select
End Function

'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''
'''''''''''''''''''''''''''''''''''''''''''''''

Private Sub AllocStructs()
    os = AllocateMem(STRUCT_SIZE)
    og = AllocateMem(STRUCT_SIZE)
    op = AllocateMem(STRUCT_SIZE)
    vc = AllocateMem(STRUCT_SIZE)
    vi = AllocateMem(STRUCT_SIZE)
    vd = AllocateMem(STRUCT_SIZE)
    vb = AllocateMem(STRUCT_SIZE)
End Sub

Private Sub FreeStructs()
    FreeMem os
    FreeMem og
    FreeMem op
    FreeMem vc
    FreeMem vi
    FreeMem vd
    FreeMem vb
End Sub

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Function WriteStream( _
    ByVal Flush As Boolean _
) As Boolean

    Dim udtPage     As ogg_page
    Dim lngRet      As Long

    Do
        If Flush Then
            lngRet = ogg_stream_flush(os.pointer, _
                                      udtPage)
        Else
            lngRet = ogg_stream_pageout(os.pointer, _
                                        udtPage)
        End If

        If lngRet = 0 Then
            Exit Do
        ElseIf lngRet < 0 Then
            Exit Function
        End If

        FileWrite hFOut, udtPage.pHeader, _
                  udtPage.header_len

        FileWrite hFOut, udtPage.pBody, _
                  udtPage.body_len
    Loop

    WriteStream = True
End Function

Private Function ProcessAudio() As Boolean
    Dim lngRet      As Long

    Do While vorbis_analysis_blockout(vd.pointer, vb.pointer) = 1
        lngRet = vorbis_analysis(vb.pointer, 0)
        If lngRet < 0 Then Exit Function

        lngRet = vorbis_bitrate_addblock(vb.pointer)
        If lngRet < 0 Then Exit Function

        Do While vorbis_bitrate_flushpacket(vd.pointer, op.pointer) > 0
            ogg_stream_packetin os.pointer, op.pointer
            If Not WriteStream(False) Then
                Exit Function
            End If
        Loop
    Loop

    ProcessAudio = True
End Function

''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''

Private Sub vorbis_info_init( _
    ByVal VorbisInfo As Long _
)

    clsVorbis.CallFunc "vorbis_info_init", _
                       VorbisInfo
End Sub

Private Sub vorbis_comment_init( _
    ByVal VorbisComment As Long _
)

    clsVorbis.CallFunc "vorbis_comment_init", _
                       VorbisComment
End Sub

Private Function vorbis_encode_init_vbr( _
    ByVal VorbisInfo As Long, _
    ByVal Channels As Long, _
    ByVal samplerate As Long, _
    ByVal base_quality As Single _
) As Long

    Dim lngQuality  As Long
    CpyMem lngQuality, base_quality, 4

    vorbis_encode_init_vbr = clsVorbis.CallFunc("vorbis_encode_init_vbr", _
                                                VorbisInfo, _
                                                Channels, _
                                                samplerate, _
                                                lngQuality)
End Function

Private Function vorbis_encode_init( _
    ByVal VorbisInfo As Long, _
    ByVal channes As Long, _
    ByVal samplerate As Long, _
    ByVal max_bitrate As Long, _
    ByVal nom_bitrate As Long, _
    ByVal min_bitrate As Long _
) As Long

    vorbis_encode_init = clsVorbis.CallFunc("vorbis_encode_init", _
                                            VorbisInfo, _
                                            channes, _
                                            samplerate, _
                                            max_bitrate, _
                                            nom_bitrate, _
                                            min_bitrate)
End Function

Private Function vorbis_analysis_init( _
    ByVal VorbisDSP As Long, _
    ByVal VorbisInfo As Long _
) As Long

    vorbis_analysis_init = clsVorbis.CallFunc("vorbis_analysis_init", _
                                              VorbisDSP, _
                                              VorbisInfo)
End Function

Private Function vorbis_block_init( _
    ByVal VorbisDSP As Long, _
    ByVal VorbisBlock As Long _
) As Long

    vorbis_block_init = clsVorbis.CallFunc("vorbis_block_init", _
                                           VorbisDSP, _
                                           VorbisBlock)
End Function

Private Sub vorbis_block_clear( _
    ByVal VorbisBlock As Long _
)
    clsVorbis.CallFunc "vorbis_block_clear", _
                       VorbisBlock
End Sub

Private Sub vorbis_dsp_clear( _
    ByVal VorbisDSP As Long _
)

    clsVorbis.CallFunc "vorbis_dsp_clear", _
                       VorbisDSP
End Sub

Private Sub vorbis_comment_clear( _
    ByVal VorbisComment As Long _
)

    clsVorbis.CallFunc "vorbis_comment_clear", _
                       VorbisComment
End Sub

Private Sub vorbis_info_clear( _
    ByVal VorbisInfo As Long _
)

    clsVorbis.CallFunc "vorbis_info_clear", _
                       VorbisInfo
End Sub

Private Sub vorbis_comment_add_tag( _
    ByVal VorbisComment As Long, _
    ByVal Tag As String, _
    ByVal TagValue As String _
)

    Dim btTag()     As Byte
    Dim btTagVal()  As Byte

    btTag = StrConv(Tag & Chr$(0), vbFromUnicode)
    btTagVal = StrConv(TagValue & Chr$(0), vbFromUnicode)

    clsVorbis.CallFunc "vorbis_comment_add_tag", _
                       VorbisComment, _
                       VarPtr(btTag(0)), _
                       VarPtr(btTagVal(0))
End Sub

Private Function vorbis_analysis_buffer( _
    ByVal VorbisDSP As Long, _
    ByVal VorbisSamples As Long _
) As Long

    vorbis_analysis_buffer = clsVorbis.CallFunc("vorbis_analysis_buffer", _
                                                VorbisDSP, _
                                                VorbisSamples)
End Function

Private Sub vorbis_analysis_wrote( _
    ByVal VorbisDSP As Long, _
    ByVal VorbisSamples As Long _
)

    clsVorbis.CallFunc "vorbis_analysis_wrote", _
                       VorbisDSP, _
                       VorbisSamples
End Sub

Private Function vorbis_analysis_headerout( _
    ByVal VorbisDSP As Long, _
    ByVal VorbisComment As Long, _
    ByVal header As Long, _
    ByVal header_comm As Long, _
    ByVal header_code As Long _
) As Long

    vorbis_analysis_headerout = clsVorbis.CallFunc("vorbis_analysis_headerout", _
                                                   VorbisDSP, _
                                                   VorbisComment, _
                                                   header, _
                                                   header_comm, _
                                                   header_code)
End Function

Private Function vorbis_analysis_blockout( _
    ByVal VorbisDSP As Long, _
    ByVal VorbisBlock As Long _
) As Long

    vorbis_analysis_blockout = clsVorbis.CallFunc("vorbis_analysis_blockout", _
                                                  VorbisDSP, _
                                                  VorbisBlock)
End Function

Private Function vorbis_analysis( _
    ByVal VorbisBlock As Long, _
    ByVal packet As Long _
) As Long

    vorbis_analysis = clsVorbis.CallFunc("vorbis_analysis", _
                                         VorbisBlock, _
                                         packet)
End Function

Private Function vorbis_bitrate_addblock( _
    ByVal VorbisBlock As Long _
) As Long

    vorbis_bitrate_addblock = clsVorbis.CallFunc("vorbis_bitrate_addblock", _
                                                 VorbisBlock)
End Function

Private Function vorbis_bitrate_flushpacket( _
    ByVal VorbisDSP As Long, _
    ByVal packet As Long _
) As Long

    vorbis_bitrate_flushpacket = clsVorbis.CallFunc("vorbis_bitrate_flushpacket", _
                                                    VorbisDSP, _
                                                    packet)
End Function

Private Function ogg_stream_init( _
    ByVal OggStream As Long, _
    ByVal SerialNo As Long _
) As Long

    ogg_stream_init = clsOgg.CallFunc("ogg_stream_init", _
                                      OggStream, _
                                      SerialNo)
End Function

Private Sub ogg_stream_clear( _
    ByVal OggStream As Long _
)

    clsOgg.CallFunc "ogg_stream_clear", OggStream
End Sub

Private Function ogg_stream_packetin( _
    ByVal OggStream As Long, _
    ByVal packet As Long _
) As Long

    ogg_stream_packetin = clsOgg.CallFunc("ogg_stream_packetin", _
                                          OggStream, _
                                          packet)
End Function

Private Function ogg_stream_flush( _
    ByVal OggStream As Long, _
    page As ogg_page _
) As Long

    ogg_stream_flush = clsOgg.CallFunc("ogg_stream_flush", _
                                       OggStream, _
                                       VarPtr(page))
End Function

Private Function ogg_stream_pageout( _
    ByVal OggStream As Long, _
    page As ogg_page _
) As Long

    ogg_stream_pageout = clsOgg.CallFunc("ogg_stream_pageout", _
                                         OggStream, _
                                         VarPtr(page))
End Function
