VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Player"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Implements IDataCallback

Private clsDSOut        As DirectSoundOut
Attribute clsDSOut.VB_VarHelpID = -1

Private clsStreams()    As ISoundStream
Private clsStream       As ISoundStream
Private lngStreamCnt    As Long
Private lngStreamIdx    As Long

Private blnEOS          As Boolean

Private lngMuteVol      As Long
Private blnMute         As Boolean

Public Event StatusChanged(ByVal Status As DSO_STATUS)
Public Event EndOfStream()

Public Property Get Mute( _
) As Boolean

    Mute = blnMute
End Property

Public Property Let Mute( _
    ByVal Value As Boolean _
)

    If Value Then
        lngMuteVol = clsDSOut.Volume
        clsDSOut.Volume = -10000
        blnMute = Value
    Else
        clsDSOut.Volume = lngMuteVol
        blnMute = Value
    End If
End Property

Public Sub AddStream( _
    stream As ISoundStream _
)

    ReDim Preserve clsStreams(lngStreamCnt) As ISoundStream
    Set clsStreams(lngStreamCnt) = stream
    lngStreamCnt = lngStreamCnt + 1
End Sub

Private Sub Class_Initialize()
    Set clsDSOut = New DirectSoundOut

    lngStreamIdx = -1
End Sub

Private Sub Class_Terminate()
    Set clsDSOut = Nothing

    If Not clsStream Is Nothing Then
        clsStream.StreamClose
    End If

    Erase clsStreams
End Sub

Public Function Play( _
) As Boolean

    Play = clsDSOut.SetStatus(DSO_PLAYING)
End Function

Public Function StopPlay( _
) As Boolean

    StopPlay = clsDSOut.SetStatus(DSO_STOPPED)
End Function

Public Function PausePlay( _
) As Boolean

    PausePlay = clsDSOut.SetStatus(DSO_PAUSING)
End Function

Public Property Get Status( _
) As DSO_STATUS

    Status = clsDSOut.Status
End Property

Public Property Get Pan( _
) As Long

    Pan = clsDSOut.Panning
End Property

Public Property Let Pan( _
    ByVal lngPan As Long _
)

    clsDSOut.Panning = lngPan
End Property

Public Property Get Volume( _
) As Long

    Volume = clsDSOut.Volume
End Property

Public Property Let Volume( _
    ByVal lngVol As Long _
)

    If blnMute Then
        lngMuteVol = lngVol
    Else
        clsDSOut.Volume = lngVol
    End If
End Property

Public Property Get positionMS( _
) As Long

    positionMS = clsDSOut.position
End Property

Public Property Get durationMS( _
) As Long

    If Not clsStream Is Nothing Then
        durationMS = clsStream.StreamInfo.Duration
    End If
End Property

Public Property Get StreamInfo( _
) As ISoundInfo

    If Not clsStream Is Nothing Then
        Set StreamInfo = clsStream.StreamInfo
    End If
End Property

Public Property Get CurrentStream( _
) As Long

    CurrentStream = lngStreamIdx
End Property

Public Property Get StreamCount( _
) As Long

    StreamCount = lngStreamCnt
End Property

Public Property Get StreamDescription( _
    ByVal Index As Long _
) As String

    StreamDescription = clsStreams(Index).Description
End Property

Public Property Get StreamExtensions( _
    ByVal Index As Long _
) As String()

    StreamExtensions = clsStreams(Index).Extensions
End Property

Public Function CaptureSamples( _
    ByVal buf_ptr As Long, _
    ByVal buf_len As Long, _
    ByVal buf_written As Long _
) As Boolean

    CaptureSamples = clsDSOut.CaptureSamples(buf_ptr, buf_len, buf_written)
End Function

Public Function QuickInfo( _
    ByVal strFile As String, _
    info As ISoundInfo _
) As SND_RESULT

    Dim strExts()   As String
    Dim strFExt     As String
    Dim strExt      As String
    Dim i           As Long

    strFExt = Mid$(strFile, InStrRev(strFile, ".") + 1)

    For i = 0 To lngStreamCnt - 1
        strExts = clsStreams(i).Extensions
        strExt = Join(strExts, ";")
        If InStr(1, strExt, strFExt, vbTextCompare) Then
            Exit For
        End If
    Next

    If i = lngStreamCnt Then
        QuickInfo = SND_ERR_INVALID_SOURCE
        Exit Function
    End If

    QuickInfo = clsStreams(i).QuickStreamInfo(strFile, info)
End Function

Public Function CloseFile( _
) As SND_RESULT

    If Not clsStream Is Nothing Then
        CloseFile = clsStream.StreamClose()
    End If
End Function

Public Function OpenFile( _
    ByVal strFile As String _
) As SND_RESULT

    Dim strExts()   As String
    Dim strFExt     As String
    Dim strExt      As String
    Dim i           As Long
    Dim sndres      As SND_RESULT

    CloseFile

    strFExt = Mid$(strFile, InStrRev(strFile, ".") + 1)

    lngStreamIdx = -1
    i = -1

Retry:
    For i = (i + 1) To lngStreamCnt - 1
        strExts = clsStreams(i).Extensions
        strExt = Join(strExts, ";")
        If InStr(1, strExt, strFExt, vbTextCompare) Then
            Exit For
        End If
    Next

    If i = lngStreamCnt Then
        OpenFile = SND_ERR_INVALID_SOURCE
        Set clsStream = Nothing
        Exit Function
    End If

    Set clsStream = clsStreams(i)

    sndres = clsStream.StreamOpen(strFile)
    If sndres = SND_ERR_SUCCESS Then
        With clsStream.StreamInfo
            If clsDSOut.InitMe(.SamplesPerSecond, .Channels, .BitsPerSample, Me) = True Then
                lngStreamIdx = i
                OpenFile = SND_ERR_SUCCESS
                modDSP.FormatChanged .SamplesPerSecond, .Channels, .BitsPerSample
            Else
                OpenFile = SND_ERR_INTERNAL
                Set clsStream = Nothing
            End If
        End With
    Else
        OpenFile = sndres

        ' maybe multiple streams have same extensions, so
        ' try to find all streams that can
        ' handle the format
        If i < lngStreamCnt Then GoTo Retry
    End If
End Function

Public Function SeekFile( _
    ByVal Value As Long, _
    ByVal Mode As SND_SEEK_MODE _
) As SND_RESULT

    Dim sndres  As SND_RESULT

    sndres = clsStream.StreamSeek(Value, Mode)

    clsDSOut.ResetBuffer

    If sndres = SND_ERR_SUCCESS Then
        clsDSOut.position = clsStream.StreamInfo.position
    End If
End Function

Private Sub IDataCallback_EndOfStream()
    RaiseEvent EndOfStream
End Sub

Private Sub IDataCallback_NeedData( _
    ByVal buf_ptr As Long, _
    ByVal buf_len As Long, _
    buf_written As Long _
)

    buf_written = 0

    If blnEOS Then Exit Sub

    If clsStream.StreamRead(buf_ptr, _
                            buf_len, _
                            buf_written) = SND_ERR_END_OF_STREAM Then

        blnEOS = True

    End If

    modDSP.ProcessData buf_ptr, buf_written
End Sub

Private Sub IDataCallback_StatusChanged( _
    ByVal stat As DSO_STATUS _
)

    If stat = DSO_STOPPED Then
        blnEOS = False
        clsStream.StreamSeek 0, SND_SEEK_PERCENT
    End If

    RaiseEvent StatusChanged(stat)
End Sub
