Attribute VB_Name = "modFunctions"
Option Explicit

' misc functions / APIs

Public Declare Function CreateWindowEx Lib "user32" _
Alias "CreateWindowExA" ( _
    ByVal dwExStyle As Long, _
    ByVal lpClassName As String, _
    ByVal lpWindowName As String, _
    ByVal dwStyle As Long, _
    ByVal x As Long, _
    ByVal y As Long, _
    ByVal nWidth As Long, _
    ByVal nHeight As Long, _
    ByVal hWndParent As Long, _
    ByVal hMenu As Long, _
    ByVal hInstance As Long, _
    ByVal lpParam As Long _
) As Long

Public Declare Function DestroyWindow Lib "user32" ( _
    ByVal hWnd As Long _
) As Long

Public Declare Function SetWindowLong Lib "user32" _
Alias "SetWindowLongA" ( _
    ByVal hWnd As Long, _
    ByVal nIndex As Long, _
    ByVal dwNewLong As Long _
) As Long

Public Declare Function GetModuleHandle Lib "kernel32" _
Alias "GetModuleHandleA" ( _
    ByVal szModule As String _
) As Long

Public Declare Function LoadLibrary Lib "kernel32" _
Alias "LoadLibraryA" ( _
    ByVal lpLibFileName As String _
) As Long

Public Declare Function FreeLibrary Lib "kernel32" ( _
    ByVal hModule As Long _
) As Long

Public Declare Function GetProcAddress Lib "kernel32" ( _
    ByVal hModule As Long, _
    ByVal lpProcName As String _
) As Long

Public Declare Function CallWindowProc Lib "user32" _
Alias "CallWindowProcA" ( _
    ByVal lpPrevWndFunc As Long, _
    ByVal hWnd As Long, _
    ByVal Msg As Long, _
    ByVal wParam As Long, _
    ByVal lParam As Long _
) As Long

Public Declare Function IsBadReadPtr Lib "kernel32" ( _
    ptr As Any, _
    ByVal ucb As Long _
) As Long

Public Declare Function IsBadWritePtr Lib "kernel32" ( _
    ptr As Any, _
    ByVal ucb As Long _
) As Long

Public Declare Function GetSystemDir Lib "kernel32" _
Alias "GetSystemDirectoryA" ( _
    ByVal lpBuffer As String, _
    ByVal nSize As Long _
) As Long

Public Declare Function GetVersionEx Lib "kernel32" _
Alias "GetVersionExA" ( _
    lpVersionInformation As OSVERSIONINFO _
) As Long

Public Declare Function GetDriveType Lib "kernel32" _
Alias "GetDriveTypeA" ( _
    ByVal nDrive As String _
) As Long

Public Declare Function GlobalUnlock Lib "kernel32.dll" ( _
    ByVal hMem As Long _
) As Long

Public Declare Function GlobalAlloc Lib "kernel32.dll" ( _
    ByVal wFlags As GMEMFlags, _
    ByVal dwBytes As Long _
) As Long

Public Declare Function GlobalFree Lib "kernel32.dll" ( _
    ByVal hMem As Long _
) As Long

Public Declare Function GlobalLock Lib "kernel32.dll" ( _
    ByVal hMem As Long _
) As Long

Public Declare Sub FillMem Lib "kernel32.dll" _
Alias "RtlFillMemory" ( _
    pData As Any, _
    ByVal dwLen As Long, _
    ByVal Fill As Byte _
)

Public Declare Sub ZeroMem Lib "kernel32.dll" _
Alias "RtlZeroMemory" ( _
    pDst As Any, _
    ByVal cBytes As Long _
)

Public Declare Sub CpyMem Lib "kernel32" _
Alias "RtlMoveMemory" ( _
    pDst As Any, _
    pSrc As Any, _
    ByVal cBytes As Long _
)


Public Enum GMEMFlags
    GMEM_FIXED = &H0
    GMEM_MOVEABLE = &H2
    GMEM_ZEROINIT = &H40
End Enum

Public Type OSVERSIONINFO
  dwOSVersionInfoSize   As Long
  dwMajorVersion        As Long
  dwMinorVersion        As Long
  dwBuildNumber         As Long
  dwPlatformId          As Long
  szCSDVersion          As String * 128
End Type

Public Type memory
    handle              As Long
    pointer             As Long
    bytes               As Long
End Type

Public Type LARGE_INTEGER
    lo                  As Long
    hi                  As Long
End Type

Public Const VER_PLATFORM_WIN32s        As Long = 0
Public Const VER_PLATFORM_WIN32_WINDOWS As Long = 1
Public Const VER_PLATFORM_WIN32_NT      As Long = 2

Public Const GWL_WNDPROC                As Long = (-4)

Private lngPower2(31)   As Long

Public Function AllocateMem( _
    ByVal bytes As Long _
) As memory

    With AllocateMem
        .bytes = bytes
        .handle = GlobalAlloc(GMEM_FIXED Or GMEM_ZEROINIT, bytes)
        .pointer = GlobalLock(.handle)
    End With
End Function

Public Sub FreeMem( _
    mem As memory _
)

    With mem
        GlobalUnlock .handle
        GlobalFree .handle

        ZeroMem mem, Len(mem)
    End With
End Sub

Public Function GetWindowsVersion() As OSVERSIONINFO
    Dim udtVersion  As OSVERSIONINFO

    udtVersion.dwOSVersionInfoSize = Len(udtVersion)
    GetVersionEx udtVersion

    GetWindowsVersion = udtVersion
End Function

Public Function GetDirLevel( _
    ByVal strFile As String _
) As Integer

    GetDirLevel = UBound(Split(strFile, "\"))
End Function

Public Function GetExtension( _
    ByVal strFile As String _
) As String

    If InStr(strFile, ".") > 0 Then
        GetExtension = Mid$(strFile, InStrRev(strFile, ".") + 1)
    Else
        GetExtension = strFile
    End If
End Function

Public Function FmtStrLen( _
    ByVal B As String, _
    Length As Integer _
) As String

    B = String(Length - Len(B), "0") & B
    FmtStrLen = B
End Function

Public Function RemNullChars( _
    ByVal strString As String _
) As String

    If InStr(strString, Chr$(0)) > 0 Then
        RemNullChars = Left$(strString, InStr(strString, Chr$(0)) - 1)
    Else
        RemNullChars = strString
    End If
End Function

Public Function IsBitSet( _
    ByVal Value As Long, _
    bit As Byte _
) As Boolean

    IsBitSet = CBool(Value And 2 ^ bit)
End Function

Public Function MKWord( _
    ByVal bH As Byte, _
    ByVal Bl As Byte _
) As Integer

    MKWord = SHL(bH, 8) Or Bl
End Function

Public Function MKDWord( _
    ByVal Wh As Integer, _
    ByVal Wl As Integer _
) As Long

    MKDWord = SHL(Wh, 16) Or Wl
End Function

Public Function LoWord( _
    ByVal DWord As Long _
) As Integer

    CpyMem LoWord, DWord, 2
    ' LoWord = DWord And &HFFFF&
End Function

Public Function HiWord( _
    ByVal DWord As Long _
) As Long

    HiWord = (DWord And &HFFFF0000) \ &H10000
End Function

Public Function LoByte( _
    ByVal word As Integer _
) As Byte

  LoByte = word And &HFF
End Function

Public Function HiByte( _
    ByVal word As Integer _
) As Byte

  HiByte = (word And &HFF00&) \ &H100
End Function

Public Function LoNibble( _
    ByVal Bt As Byte _
) As Byte

    LoNibble = Bt And &HF
End Function

Public Function HiNibble( _
    ByVal Bt As Byte _
) As Byte

    HiNibble = (Bt And &HF0) \ &H10
End Function

' >> Operator
' by VB-Accelerator
Public Function SHR( _
    ByVal lThis As Long, _
    ByVal lBits As Long _
) As Long

    Static Init As Boolean

    If Not Init Then InitShifting: Init = True

    If (lBits <= 0) Then
        SHR = lThis
    ElseIf (lBits > 63) Then
        Exit Function
    ElseIf (lBits > 31) Then
        SHR = 0
    Else
        If (lThis And lngPower2(31)) = lngPower2(31) Then
            SHR = (lThis And &H7FFFFFFF) \ lngPower2(lBits) Or lngPower2(31 - lBits)
        Else
            SHR = lThis \ lngPower2(lBits)
        End If
    End If

End Function

' << Operator
' by VB-Accelerator
Public Function SHL( _
    ByVal lThis As Long, _
    ByVal lBits As Long _
) As Long

    Static Init As Boolean

    If Not Init Then InitShifting: Init = True

    If (lBits <= 0) Then
        SHL = lThis
    ElseIf (lBits > 63) Then
        Exit Function
    ElseIf (lBits > 31) Then
        SHL = 0
    Else
        If (lThis And lngPower2(31 - lBits)) = lngPower2(31 - lBits) Then
            SHL = (lThis And (lngPower2(31 - lBits) - 1)) * lngPower2(lBits) Or lngPower2(31)
        Else
            SHL = (lThis And (lngPower2(31 - lBits) - 1)) * lngPower2(lBits)
        End If
    End If

End Function

' power of 2
Private Sub InitShifting()
    Dim i   As Long
    For i = 0 To 30: lngPower2(i) = 2& ^ i: Next
    lngPower2(31) = &H80000000
End Sub

Public Function Dbl2LargeInt( _
    ByVal dbl As Double _
) As LARGE_INTEGER

    Dim strHex  As String
    Dim qwRet   As LARGE_INTEGER

    strHex = FmtStrLen(DecToHex(dbl), 16)

    ' lo DWORD
    qwRet.lo = Val("&H" & Right$(strHex, 8) & "&")
    ' hi DWORD
    qwRet.hi = Val("&H" & Left$(strHex, 8) & "&")

    Dbl2LargeInt = qwRet
End Function

' http://pscode.com/vb/scripts/ShowCode.asp?txtCodeId=49987&lngWId=1
Public Function DecToHex( _
    ByVal nSource As Double _
) As String

    Const BASECHAR  As String = "0123456789ABCDEF"
    Dim n           As Double
    Dim nSrc        As Double

    On Error GoTo ErrorLine

    If (nSource = 0) Then
        DecToHex = "00"
        Exit Function
    End If

    If (nSource < 2147483648#) Then
        DecToHex = Hex(nSource)
    Else
        nSrc = nSource

        Do
            n = CDec(nSrc - (16 * Int(nSrc / 16)))
            DecToHex = Mid$(BASECHAR, n + 1, 1) & DecToHex
            nSrc = CDec(Int(nSrc / 16))
        Loop While (nSrc > 0)

    End If

    If (Len(DecToHex) Mod 2) Then
        DecToHex = "0" & DecToHex
    End If
    Exit Function

ErrorLine:
    DecToHex = ""
End Function

Public Function GetSystemDirectory( _
) As String

    Dim strTemp As String

    strTemp = Space$(256)
    strTemp = Left$(strTemp, GetSystemDir(strTemp, _
                                          Len(strTemp)))

    If Right$(strTemp, 1) <> "\" Then strTemp = strTemp & "\"
    GetSystemDirectory = strTemp
End Function

Public Function TrimEx( _
    ByVal strVal As String _
) As String

    TrimEx = Trim$(Left$(strVal, InStr(strVal, Chr$(0)) - 1))
End Function

Public Function Ptr2Str( _
    ByVal ptr As Long _
) As String

    Dim strBuffer   As String
    Dim btChr       As Byte
    Dim lngPos      As Long

    If ptr = 0 Then Exit Function

    btChr = 1
    Do While btChr > 0
        CpyMem btChr, ByVal ptr + lngPos, 1
        If btChr > 0 Then
            strBuffer = strBuffer & Chr$(btChr)
        End If
        lngPos = lngPos + 1
    Loop

    Ptr2Str = strBuffer
End Function

Public Function FmtTime( _
    ByVal lngMS As Long _
) As String

    Dim lngMin  As Long
    Dim lngSec  As Long

    lngSec = lngMS / 1000
    lngMin = lngSec \ 60
    lngSec = lngSec Mod 60

    FmtTime = lngMin & ":" & format(lngSec, "00")
End Function

Public Function FormatFileSize( _
    ByVal dblFileSize As Double, _
    Optional ByVal strFormatMask As String _
) As String

    Select Case dblFileSize
        Case 0 To 1023 ' Bytes
            FormatFileSize = format(dblFileSize) & " bytes"

        Case 1024 To 1048575 ' KB
            If strFormatMask = Empty Then strFormatMask = "###0"
            FormatFileSize = format(dblFileSize / 1024#, strFormatMask) & " KB"

        Case 1024# ^ 2 To 1073741823 ' MB
            If strFormatMask = Empty Then strFormatMask = "###0.0"
            FormatFileSize = format(dblFileSize / (1024# ^ 2), strFormatMask) & " MB"

        Case Is > 1073741823# ' GB
            If strFormatMask = Empty Then strFormatMask = "###0.0"
            FormatFileSize = format(dblFileSize / (1024# ^ 3), strFormatMask) & " GB"
    End Select
End Function

Public Function GetFilePath( _
    ByVal strPath As String _
) As String

    Dim strParts()  As String

    strParts = Split(strPath, "\")
    GetFilePath = Left$(strPath, Len(strPath) - Len(strParts(UBound(strParts))))
End Function

Public Function GetFileExt( _
    ByVal strPath As String _
) As String

    GetFileExt = Mid$(strPath, InStrRev(strPath, ".") + 1)
End Function

Public Function GetFileWOExt( _
    ByVal strPath As String _
) As String

    If InStrRev(strPath, ".") > 0 Then
        strPath = Left$(strPath, InStrRev(strPath, "."))
    End If

    GetFileWOExt = strPath
End Function

Public Function AddSlash( _
    ByVal strPath As String _
) As String

    AddSlash = strPath & IIf(Right$(strPath, 1) = "\", "", "\")
End Function

Public Function GetFilename( _
    ByVal strPath As String _
) As String

    GetFilename = Mid$(strPath, InStrRev(strPath, "\") + 1)
End Function

Public Function BtArrToHex( _
    btArr() As Byte _
) As String

    Dim strHex  As String
    Dim i       As Long

    For i = LBound(btArr) To UBound(btArr)
        strHex = strHex & FmtStrLen(Hex$(btArr(i)), 2)
    Next

    BtArrToHex = strHex
End Function

Public Function StrToHex( _
    strValue As String _
) As String

    Dim strHex  As String
    Dim i       As Long

    For i = 1 To Len(strValue)
        strHex = strHex & FmtStrLen(Hex$(Asc(Mid$(strValue, i, 1))), 2)
    Next

    StrToHex = strHex
End Function

Public Function HexToStr( _
    strHex As String _
) As Byte()

    Dim i       As Long
    Dim strOut  As String

    For i = 1 To Len(strHex) - 1 Step 2
        strOut = strOut & Chr$(Val("&H" & Mid$(strHex, i, 2)))
    Next

    HexToStr = strOut
End Function

Public Function HexToBtArr( _
    strHex As String _
) As Byte()

    Dim i       As Long
    Dim btOut() As Byte

    ReDim btOut(Len(strHex) / 2 - 1) As Byte

    For i = 1 To Len(strHex) - 1 Step 2
        btOut(i / 2 - 0.5) = Val("&H" & Mid$(strHex, i, 2))
    Next

    HexToBtArr = btOut
End Function
